/*
OpenLaserShowController.cpp : Defines the initialization routines for the DLL.

Copyright (c) 2009 Chris Favreau

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include "stdafx.h"
#include "OpenLaserShowController.h"
#include "LaserOutputDlg.h"
#include <direct.h>				// _getcwd(), ... 

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

//
//TODO: If this DLL is dynamically linked against the MFC DLLs,
//		any functions exported from this DLL which call into
//		MFC must have the AFX_MANAGE_STATE macro added at the
//		very beginning of the function.
//
//		For example:
//
//		extern "C" BOOL PASCAL EXPORT ExportedFunction()
//		{
//			AFX_MANAGE_STATE(AfxGetStaticModuleState());
//			// normal function body here
//		}
//
//		It is very important that this macro appear in each
//		function, prior to any calls into MFC.  This means that
//		it must appear as the first statement within the 
//		function, even before any object variable declarations
//		as their constructors may generate calls into the MFC
//		DLL.
//
//		Please see MFC Technical Notes 33 and 58 for additional
//		details.
//

// Laser Output Dialog
bool g_bInitialized = false;
CLaserOutputDlg *g_pOutput = NULL;
int g_iMaxDevices = 1;

// A WinApp is needed for the CDialog window we use in this test application
CWinApp theApp;

// TODO - move this function elsewhere
void ReadSettingsFromFile()
{
	char file_name[_MAX_PATH];		// Name of the INI file... with path..
	char working_dir[_MAX_PATH];	// Current Working Directory

	// Open up the INI file for configuration
	// Get the current working directory
	_getcwd(working_dir, _MAX_PATH);
	// Attach the INI filename
	sprintf_s(file_name, _MAX_PATH, "%s\\%s", working_dir, "olsc_demo.ini");

	// The number of supported devices
	g_iMaxDevices = GetPrivateProfileIntA("OLSC DEMO", "DEVICE COUNT", 2, file_name);
}

#define OPEN_LASER_SHOW_DEVICE_API_VERSION	2

#define OPEN_LASER_SHOW_DEVICE_NAME "Simulation Display"

// OLSC_GetAPIVersion
// Inputs: None
// Outputs: Returns a single integer that represents the version number
// Description: Returns the version of the OpenDAC API
//				so we don't make any critical errors when using this DLL
//
OLSC_API int WINAPI OLSC_GetAPIVersion(void)
{
	return OPEN_LASER_SHOW_DEVICE_API_VERSION;
}

// OLSC_GetAPIVersion
// Inputs: Pointer to a string at least 64 characters in length
// Outputs: Returns success or failure
// Description: Returns the string name of the OpenDC Interface
//
int WINAPI OLSC_GetInterfaceName(char *pString)
{
	sprintf(pString, OPEN_LASER_SHOW_DEVICE_NAME);

	return 1;
}

// OLSC_Initialize
// Inputs: None
// Outputs: Returns the number of available devices.
// Description: Initializes the hardware
int WINAPI OLSC_Initialize(void)
{
	if (!g_bInitialized)
	{
		AFX_MANAGE_STATE(AfxGetStaticModuleState());

		// Read the settings from the file (OLSC_DEMO.INI)
		ReadSettingsFromFile();

		// Create a new laser output dialog window
		g_pOutput = new CLaserOutputDlg;
		
		// Pass it the number of Vector Controls we want (= to the number of devices we were configured for)
		g_pOutput->SetDisplayCount(g_iMaxDevices);

		// Now create the window and show it
		g_pOutput->Create(IDD_LASERVIEW_DIALOG, NULL);
		g_pOutput->ShowWindow(SW_SHOW);

		// Set the initialized flag
		g_bInitialized = true;
	}

	// Return 0 for no devices available
	// Return the number of devices available
	return g_iMaxDevices;
}

// OLSC_Shutdown
// Inputs: None
// Outputs: Returns success or failure
// Description: Shuts down all devices
//
int WINAPI OLSC_Shutdown(void)
{
	// Check to make sure we have been initialized
	if (!g_bInitialized) return 0;

	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	// Destroy the window
	delete g_pOutput;
	g_pOutput = NULL;
	// Reset the intialized flag
	g_bInitialized = false;

	// Return 0 for failure
	// Return > 0 for success
	return TRUE;
}

// OLSC_GetDeviceCount
// Inputs: None
// Outputs: Returns the number of available devices
// Description: Has to be called AFTER Intialize has been called.  Thsi function
// is to be used to query the number of available devices without calling
// intialize again and again
int WINAPI OLSC_GetDeviceCount(void)
{
	// Check to make sure we have been initialized
	if (!g_bInitialized) return 0;

	// Return the maximum number of devices we support
	return g_iMaxDevices;
}

// OLSC_GetCapabilities
// Inputs:  device Number, device capabilties structure to be filled in
// Outputs: Returns success or failure
// Description: Gets the capabilties of each device attached
//				Caps:
//				Device Name
//				Device Version
//				Min Speed/Max Speed
//				Min Frame Size/Max Frame Size
//				DMX In/Out?
//				TTL In/Out?
//				X/Y Resolution (8 to 16 bits)
//				Color Resolution (1 to 16 bits)
//				Uses callbacks?
//				Some reserved space for future use
int WINAPI OLSC_GetDeviceCapabilities(int device_number, struct LASER_SHOW_DEVICE_CAPABILITIES &device_capabilities)
{
	// Check to make sure we have been initialized
	if (!g_bInitialized) return 0;

	// Check the device number (this function cannot take device 0 as this means all devices)
	if (device_number < 1) return 0;
	if (device_number > g_iMaxDevices) return 0;

	// Fill out the device capabilities

	// These capabilities describe the features of the interface
	// I.E. what is available and not available to use
	device_capabilities.has_dmx_in = false;
	device_capabilities.has_dmx_out = false;
	device_capabilities.has_ttl_in = false;
	device_capabilities.has_ttl_out = false;
	device_capabilities.max_frame_size = MAX_FRAME_SIZE;
	device_capabilities.max_speed = 30000;
	device_capabilities.min_frame_size = 1;
	device_capabilities.min_speed =  1;
	device_capabilities.uses_callbacks = false;
	device_capabilities.version_major = 1;
	device_capabilities.version_minor = 1;

	// These capabilities are purely for information only
	// All values passed to the OLSC interface are to be full size unsigned 16 bit
	device_capabilities.color_resolution = 8;
	device_capabilities.xy_resolution = 16;
	_snprintf(device_capabilities.name, 1024, "Laser Display %d", device_number);

	// Return 0 for failure
	// Return > 0 for success
	return 1;
}

// OLSC_SetConfigFile
// Inputs: configuration path string (expected to have leading '/')
// Outputs: Returns success or failure
// Description: Sets the path name to the configuration file for this OLSC interface
// Other supporting files such as libraries and other configuration files must be at this location too
int WINAPI OLSC_SetConfigFilePath(char *pPathString)
{
	// Check to see if the path pointer is good
	if (!pPathString) return 0;

	// TODO - load the configuration file using this path

	// Return 0 for failure
	// Return > 0 for success
	return 1;
}

// OLSC_ShowConfigWindow
// Inputs: none
// Outputs: success or failure
// Description:  Calling this function tells the OLSC interface to show its built in configuration window if it has one
// a return of failure indicates that no window is available or could not be opened.  Configuration
// windows are expected to NOT be blocking (Modal).
int WINAPI OLSC_ShowConfigWindow(void)
{
	// TODO - create a configuration window

	// Return that there is no configuration window available for this interface YET.
	return 0;
}

// OLSC_GetLastErrorNumber
// Inputs: device number, error string pointer, and string length max.
// Outputs: error number, and actual error string
// Description: Get the string and/or number associated with the last error
//
int WINAPI OLSC_GetLastErrorNumber(int device_number, int &number, char *string_pointer, int string_length)
{
	// Check to make sure we have been initialized
	if (!g_bInitialized) return 0;

	// Check the device number (this function cannot take device 0 as this means all devices)
	if (device_number < 0) return 0;
	if (device_number > g_iMaxDevices) return 0;
	
	// Return 0 for failure
	// Return > 0 for success
	return 1;
}

// OLSC_Play
// Inputs: device number
// Outputs: Returns success or failure
// Description: Starts the output on a particular device or all devices
//
int WINAPI OLSC_Play(int device_number)
{
	// Check to make sure we have been initialized
	if (!g_bInitialized) return 0;

	// Check the device number
	if (device_number < 0) return 0;
	if (device_number > g_iMaxDevices) return 0;

	// Device 0 means all devices

	// TODO - display the current frame in our buffer (we may have been blanked)

	// Return 0 for failure
	// Return > 0 for success
	return 1;
}

// OLSC_Pause
// Inputs: device number
// Outputs: Returns success or failure
// Description: Stops the output on a particular device or all devices
//
int WINAPI OLSC_Pause(int device_number)
{
	// Check to make sure we have been initialized
	if (!g_bInitialized) return 0;

	// Check the device number
	if (device_number < 0) return 0;
	if (device_number > g_iMaxDevices) return 0;

	// Device 0 means all devices

	// TODO - stop displaying the frame and blank the display

	// Return 0 for failure
	// Return > 0 for success
	return 1;
}

// OLSC_Shutter
// Inputs: device number, shutter state (LASER_SHOW_DEVICE_SHUTTER_STATE)
// Outputs: Returns success or failure
// Description: Turns the laser show device's shutter On/Off
//
int WINAPI OLSC_Shutter(int device_number, int state)
{
	// Check to make sure we have been initialized
	if (!g_bInitialized) return 0;

	// Check the device number
	if (device_number < 0) return 0;
	if (device_number > g_iMaxDevices) return 0;

	// Device 0 = all initialized devices, other wise address the individual device

	// Return 0 for failure
	// Return > 0 for success
	return 1;
}

// OLSC_WriteFrame
// Inputs: device number, frame structure (LASER_SHOW_DEIVCE_FRAME)
// Outputs: Returns success or failure
// Description:
// Writes a frame to a particular device or all devices
// Pass in a frame:
// Point Count
// Display Speed PPS
// Point Array:
// X 16 bit unsigned
// Y 16 bit unsigned
// R 16 bit unsigned
// G 16 bit unsigned
// B 16 bit unsigned
// I 16 bit unsigned
//
// ** Any frame that is written will be displayed until another frame is written to take its place or the Output is Paused
//
int WINAPI OLSC_WriteFrame(int device_number, struct LASER_SHOW_DEVICE_FRAME frame)
{
	// Check to make sure we have been initialized
	if (!g_bInitialized) return 0;

	// Check the device number
	if (device_number < 0) return 0;
	if (device_number > g_iMaxDevices) return 0;

	// TODO - 
	// Device 0 means all devices otherwise just address one display

	// TODO - put this frame in a lockless queue so that we may display the frame only when ready
	// which would be when the previous frame has finished display ... which would be
	// when the number of points in this frame has been converted to time (using the frames PPS)
	// and that time has expired.

	// Write the frame to the dialog window
	g_pOutput->WriteFrame(frame);

	// Return 0 for failure
	// Return > 0 for success
	return 1;
}

// OLSC_WritePoint
// Inputs: device number, point
// Outputs: success or failure
// Description: Stores a point in a buffer to be written later
//
int WINAPI OLSC_WritePoint(int device_number, struct LASER_SHOW_DEVICE_POINT point)
{
	// Return 0 for failure
	// Return > 0 for success
	return 1;
}

// OLSC_WriteBufferedFrame
// Inputs: device number
// Outputs: success or failure
// Description: Writes all of the points that have been buffered up to this point
//
int WINAPI OLSC_WriteBufferedFrame(int device_number)
{
	// Return 0 for failure
	// Return > 0 for success
	return 1;
}

// OLSC_GetStatus()
// Inputs: device number, status DWORD
// Outputs:  Returns success or failure
// Description:
// Gets the status of a particular device
// Can be used for polling or in conjunction with Windows Message Callback
// Status Structure or Bit Mask:
// Bit 0 = Buffer Full
// Bit 1 = Buffer Empty
// Bit 2 = DMX Out Complete
// Bit 3 = DMX In Ready
// Bit 4 = TTL Out Complete
// Bit 5 = TLL In Ready
// Bit 6 = Resreved
// Bit 7 = Reserved
// Bit 8 = Reserved
// Bit 9-30 = Reserved
// Bit 31 = An Error Occured
int WINAPI OLSC_GetStatus(int device_number, DWORD &status)
{
	// Check to make sure we have been initialized
	if (!g_bInitialized) return 0;

	// Check the device number (this function cannot take device 0 as this means all devices)
	if (device_number < 0) return 0;
	if (device_number > g_iMaxDevices) return 0;
	
	// TODO - implement a multimedia timer to time frames and possible call the callback function
	// This should toggle the Buffer Full and Buffer Empty bits so that an application that is polling
	// will be able to tell when we are done displaying.

	// Return 0 for failure
	// Return > 0 for success
	return 1;
}

// OLSC_SetCallback()
// Inputs: device number, parent window handle (for SendMessage), message DWORD
// Outputs:  Returns success or failure
// Description:
// Sets the call back window parent handle and windows message to send to run the callback
// and set the call back notification mask bits per the GetStatus mask.  The Callback message
// will return the Device Number in the wParam and the Status mask in the lParam.
// We all work in windows right?
// This interface is optional in the DLL and in the application.  The DLL writers should take this into account.
// So if you use the callback ... be able to function without it... it is used for efficiency.
//int WINAPI OLSC_SetCallback(int device_number, HWND parent_window_handle, DWORD message)
OLSC_API int WINAPI OLSC_SetCallback(int device_number, POLSCCALLBACK pCallback)
{
	// Check to make sure we have been initialized
	if (!g_bInitialized) return 0;

	// Check the device number (this function cannot take device 0 as this means all devices)
	if (device_number < 0) return 0;
	if (device_number > g_iMaxDevices) return 0;
	
	// TODO - implement callbacks using the mmtimer we mention in the GetStatus function.

	// TODO - decide if we should use windows messages or function pointers for the callback.
	// This library is for Windows only right?

	// Return 0 for failure
	// Return > 0 for success
	return 0;
}

// OLSC_WriteDMX()
// Inputs: device number, start address, data pointer, data length
// Outputs:  Returns success or failure
// Description:
// Writes DMX data to a particular device
int WINAPI OLSC_WriteDMX(int device_number, int start_address, unsigned char *data_pointer, int length)
{
	// TODO - develop a small section that displays the DMX output

	// Return 0 for failure
	// Return > 0 for success
	return 0;
}

// OLSC_ReadDMX()
// Inputs: device number, start address, data pointer, data length
// Outputs:  Returns success or failure
// Description:
// Reads DMX data from a particular device
int WINAPI OLSC_ReadDMX(int device_number, int start_address, unsigned char *data_pointer, int length)
{
	// TODO - readout what we put in

	// Return 0 for failure
	// Return > 0 for success
	return 0;
}

// OLSC_WriteTTL()
// Inputs: device number, data DWORD (up to 32 bits of TTL outputs)
// Outputs:  Returns success or failure
// Description:
// Writes TTL data from a particular device
int WINAPI OLSC_WriteTTL(int device_number, DWORD data)
{
	// TODO - develop a small section for 8 bits of TTL output

	// Return 0 for failure
	// Return > 0 for success
	return 0;
}

// OLSC_ReadTTL()
// Inputs: device number, data DWORD (up to 32 bits of TTL inputs)
// Outputs:  Returns success or failure
// Description:
// Reads TTL data from a particular device
int WINAPI OLSC_ReadTTL(int device_number, DWORD &data)
{
	// TODO - develop a small section in the display with 8 bits worth of check boxes to simulate inputs

	// Return 0 for failure
	// Return > 0 for success
	return 0;
}
