/*
**********************************************
** Vector Display Control
**
***********************************************

Copyright (c) 2009 Chris Favreau

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/

#include "stdafx.h"
#include "VectorDisplayCtrl.h"

BEGIN_MESSAGE_MAP(CVectorDisplayCtrl, CWnd)
	//{{AFX_MSG_MAP(CVectorDisplayCtrl)
	ON_WM_SIZE()
	ON_WM_PAINT()
	ON_WM_ERASEBKGND()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

CVectorDisplayCtrl::CVectorDisplayCtrl()
{
	// Initialize the frame
	m_frame.points = NULL;
	m_frame.display_speed = 0;
	m_frame.point_count = 0;

	// Initialize the Control
	Init();
}

CVectorDisplayCtrl::~CVectorDisplayCtrl()
{
	// ReInitialize the Control
	Init();

	// Delete the frame buffer
	delete [] m_frame.points;
	m_frame.points = NULL;
}

BOOL CVectorDisplayCtrl::Create(DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID) 
{
	static CString className = AfxRegisterWndClass(CS_HREDRAW | CS_VREDRAW) ;

	BOOL bRetv = CWnd::CreateEx(/*WS_EX_CLIENTEDGE | WS_EX_STATICEDGE*/0, className, NULL, dwStyle, rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top, pParentWnd->GetSafeHwnd(), 0);

	return bRetv;
}

void CVectorDisplayCtrl::Init()
{
	// Initialize the internal variables
	m_bDisplayFrame = true;
	m_bEraseBackground = true;
	dXPointsPerPixel = 0.0;
	dYPointsPerPixel = 0.0;
	
	m_rectDisplay.bottom = 0;
	m_rectDisplay.top = 0;
	m_rectDisplay.left = 0;
	m_rectDisplay.right = 0;

	// Create our internal frame buffer from the LASER_SHOW_DEVICE_FRAME structure
	m_frame.point_count = 0;
	if (m_frame.points) delete [] m_frame.points;
	m_frame.points = new LASER_SHOW_DEVICE_POINT[MAX_FRAME_SIZE];
	m_frame.display_speed = 0;
}

void CVectorDisplayCtrl::SetPointList(LASER_SHOW_DEVICE_FRAME frame)
{
	// Stop displaying the frame
	m_bDisplayFrame = false;

	// Get the size of the frame in points
	int points = frame.point_count;
	// Bounds check the frame
	if (points > MAX_FRAME_SIZE) points = MAX_FRAME_SIZE;
	if (points < 0) return;

	// Make a copy of the frame
	m_frame.display_speed = frame.display_speed;
	memcpy(m_frame.points, frame.points, sizeof(LASER_SHOW_DEVICE_POINT) * points);
	m_frame.point_count = points;

	// Display the frame now
	m_bDisplayFrame = true;

	// Cause a redraw
	Invalidate();
}

void CVectorDisplayCtrl::SetRectangle(CRect control_rect)
{
	// Set our display rectangle
	m_rectDisplay = control_rect;
}

void CVectorDisplayCtrl::DrawDisplay(CDC *pDC)
{
	if (pDC == NULL)
		return;

	if (!m_bDisplayFrame)
		return;

	// Clear the background
	pDC->FillSolidRect(&m_rectDisplay,RGB(0,0,0));

	// Create our first pen
	CPen first_pen;
	first_pen.CreatePen(PS_SOLID, 1, RGB(m_frame.points[0].r, m_frame.points[0].g, m_frame.points[0].b));

	// Setup the Right Foreground Color
	CPen *pOldPen = pDC->SelectObject(&first_pen);
	
	// Calculate Point Display Conversion (for a 16 bit X,Y location)
	dXPointsPerPixel = (double)m_rectDisplay.Width() / 65535.0;
	dYPointsPerPixel = (double)m_rectDisplay.Height() / 65535.0;

	// Draw the first point
	int point_counter = 0;

	double dXPixel,dYPixel;
	int XPixel,YPixel;
	CRect point_rect;

	dXPixel = (int)m_frame.points[point_counter].x;
	dXPixel *= dXPointsPerPixel;

	// TODO - fix the inversion issue with the X axis
	XPixel = ((int)dXPixel + m_rectDisplay.left);
	//XPixel = m_rectDisplay.Width() - ((int)dXPixel + m_rectDisplay.left);

	dYPixel = (int)m_frame.points[point_counter].y;
	dYPixel *= dYPointsPerPixel;
	YPixel = m_rectDisplay.Height() - ((int)dYPixel + m_rectDisplay.top);

	pDC->MoveTo(XPixel,YPixel);

	point_rect.left = XPixel - 2;
	point_rect.top = YPixel - 2;
	point_rect.right = XPixel + 2;
	point_rect.bottom = YPixel + 2;
	//pDC->Rectangle(&point_rect);

	// Draw the rest of the points
	for (;point_counter < m_frame.point_count; point_counter++)
	{
		// Create a pen to suit the current point
		CPen current_pen;
		current_pen.CreatePen(PS_SOLID, 1, RGB(m_frame.points[point_counter].r, m_frame.points[point_counter].g, m_frame.points[point_counter].b));

		// Select the pen into the GDI
		pDC->SelectObject(&current_pen);

		// Calculate the next point
		dXPixel = m_frame.points[point_counter].x;
		dXPixel *= dXPointsPerPixel;
		// TODO -fix the inversion issue with the X axis
		XPixel = ((int)dXPixel + m_rectDisplay.left);
		//XPixel = m_rectDisplay.Width() - ((int)dXPixel + m_rectDisplay.left);
		dYPixel = m_frame.points[point_counter].y;
		dYPixel *= dYPointsPerPixel;
		YPixel = m_rectDisplay.Height() - ((int)dYPixel + m_rectDisplay.top);

		point_rect.left = XPixel - 2;
		point_rect.top = YPixel - 2;
		point_rect.right = XPixel + 2;
		point_rect.bottom = YPixel + 2;
		//pDC->Rectangle(&point_rect);
		
		// Draw the line from the last point
		pDC->LineTo(XPixel, YPixel);
	}

	// Free up the Display Pen
	pDC->SelectObject(pOldPen);
}

void CVectorDisplayCtrl::OnSize(UINT nType, int cx, int cy) 
{
	/*
	if (m_hWnd)
	{
		CWnd::OnSize(nType, cx, cy) ;
		GetClientRect(&m_rcClient) ;

		// set some member variables to avoid multiple function calls
		m_nClientHeight = m_rcClient.Height() ;
		m_nClientWidth  = m_rcClient.Width() ;

		InitializeDibComponents(m_hWnd);
	
		Clear();
	}
	*/
}

void CVectorDisplayCtrl::OnPaint() 
{
	// Prevent the background from being erased while the window is being drawn to
	m_bEraseBackground = false;

	CPaintDC dc(this) ;  // device context for painting

	DrawDisplay(&dc);

	m_bEraseBackground = true;
}

BOOL CVectorDisplayCtrl::OnEraseBkgnd(CDC* pDC) 
{
	if (m_bEraseBackground)
		return CWnd::OnEraseBkgnd(pDC);
	else
		return FALSE;
}

